/*
 * Transitive closure in memory for un-weighted, directed graph.
 * 
 * The matrix will be N by N where N is the highest numbered node.
 * 
 * 0 is used as no path, all weights are equal to 1.
 * 
 */

import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;

import org.apache.cassandra.thrift.*;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.transport.TFramedTransport;
import org.apache.thrift.transport.TTransport;
import org.apache.thrift.transport.TSocket;


public class TC_inMem_D {
	public static void main(String[] args) throws Exception {
		// Set up decoder.
		Charset charset = Charset.forName("UTF-8");
		CharsetDecoder decoder = charset.newDecoder();

		// Set up connection.
		TTransport transport = new TFramedTransport(new TSocket("10.9.73.119", 9160));
		TProtocol protocol = new TBinaryProtocol(transport);
		Cassandra.Client client = new Cassandra.Client(protocol);
		transport.open();

		// Set the keyspace we are using.
		client.set_keyspace("CDM");

		// Need to find the highest numbered node	
		int high_node = 0;

		ByteBuffer query = ByteBuffer.wrap("SELECT key FROM Nodes;".getBytes());
		CqlResult results = client.execute_cql_query(query, Compression.NONE);

		// Go through the results and find the highest numbered node
		for (CqlRow row : results.getRows()) {
			for (Column col : row.getColumns()) {
				String name = decoder.decode(col.name).toString();
				int value = Integer.parseInt(decoder.decode(col.value).toString());
				if (name.contentEquals("key")) {
					if (value > high_node) high_node = value;
				}
			}
		}
		
		int[][] node_list = new int[high_node][high_node];
		int start_node = 0;
		int end_node = 0;
		
		// Initialize, we can always get to ourself
		for (int i = 0; i < high_node; ++i) {
				node_list[i][i] = 1;
		}
		
		// Set the edege query and execute.
		query = ByteBuffer.wrap("SELECT start, end FROM Edges;".getBytes());
		results = client.execute_cql_query(query, Compression.NONE);
		
		// Go through all the edges and set the weights equal to 1
		for (CqlRow row : results.getRows()) {
			for (Column col : row.getColumns()) {
				if (decoder.decode(col.name).toString().contentEquals("start")) {
					start_node = Integer.parseInt(decoder.decode(col.value).toString()) - 1;
				} else {
					end_node = Integer.parseInt(decoder.decode(col.value).toString()) - 1;
				}
			}
			node_list[start_node][end_node] = 1;
			
		}
				
		transport.flush();
		transport.close();
		
		// Print out what we've got.		
		System.out.print("\t");
		for (int i = 0; i < high_node; ++i) {
			System.out.print("\t" + (i+1) + ":");
		}
		System.out.print("\n");
		
		for (int i = 0; i < node_list.length; ++i ) {
			System.out.print((i + 1) + "\t|");
			for (int j = 0; j < node_list.length; ++j) {
				System.out.print("\t" + node_list[i][j]);
			}
			System.out.print("\n");
		}
			
		// Calculate the transitive closure.
		
		/*
		 * All paths are equal in this implementation.
		 * Paths are directed.
		 * 0 = no path.
		 */
		for (int k = 0; k < high_node; ++k) {
			for (int i = 0; i < high_node; ++i) {
				for (int j = 0; j < high_node; ++j) {
					if (node_list[i][k] + node_list[k][j] == 2) { // This is a path from [i][k] to [k][j]
						node_list[i][j] = 1;
					}
				}
			}
		}
			
		// Print out the results of the closure.
		System.out.print("\n\t");
		for (int i = 0; i < high_node; ++i) {
			System.out.print("\t" + (i + 1) + ":");
		}
		System.out.print("\n");
		
		for (int i = 0; i < node_list.length; ++i ) {
			System.out.print((i+1) + "\t|");
			for (int j = 0; j < node_list.length; ++j) {
				System.out.print("\t" + node_list[i][j]);
			}
			System.out.print("\n");
		}
		

	}
}
