/*
 * Transitive closure in memory for weighted, undirected graph.
 * 
 * The matrix will be N by N where N is the highest numbered node.
 * 
 * 0 is used as no path, so the self paths will need to be set to very small not 0 number.
 * 
 */

import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.util.List;

import org.apache.cassandra.thrift.*;
import org.apache.thrift.protocol.TProtocol;
import org.apache.thrift.transport.TFramedTransport;
import org.apache.thrift.transport.TTransport;
import org.apache.thrift.transport.TSocket;


public class TC_inMem_W {
	public static void main(String[] args) throws Exception {
		// Set up decoder.
		Charset charset = Charset.forName("UTF-8");
		CharsetDecoder decoder = charset.newDecoder();

		// Set up connection.
		TTransport transport = new TFramedTransport(new TSocket("10.9.73.119", 9160));
		TProtocol protocol = new TBinaryProtocol(transport);
		Cassandra.Client client = new Cassandra.Client(protocol);
		transport.open();

		// Set the keyspace we are using.
		client.set_keyspace("CDM");

		// Need to find the highest numbered node	
		int high_node = 0;

		ByteBuffer query = ByteBuffer.wrap("SELECT key FROM Nodes;".getBytes());
		CqlResult results = client.execute_cql_query(query, Compression.NONE);

		// Go through the results and find the highest numbered node
		for (CqlRow row : results.getRows()) {
			for (Column col : row.getColumns()) {
				String name = decoder.decode(col.name).toString();
				int value = Integer.parseInt(decoder.decode(col.value).toString());
				if (name.contentEquals("key")) {
					if (value > high_node) high_node = value;
				}
			}
		}

		// Set up the edge matrix
		double[][] node_list = new double[high_node][high_node];
		int start_node = 0;
		int end_node = 0;
		double weight = 0;

		// Initialize, we can always get to ourself
		for (int i = 0; i < high_node; ++i) {
			node_list[i][i] = 0.00001; // 0 is being used as infinity. This needs to be smaller than any of the weights.
		}

		// Set the edge query and execute.
		query = ByteBuffer.wrap("SELECT start, end, weight FROM WEdges;".getBytes());
		results = client.execute_cql_query(query, Compression.NONE);

		// Go through all the edges and set the weights
		List<CqlRow> rows = results.getRows();
		for (CqlRow row : rows) {
			List<Column> columns = row.getColumns();
			for (Column col : columns) {
				String colName = decoder.decode(col.name).toString();
				String colValue = decoder.decode(col.value).toString();

				if (colName.contentEquals("start")) {
					start_node = Integer.parseInt(colValue) - 1;
				}
				if (colName.contentEquals("end")){
					end_node = Integer.parseInt(colValue) - 1;
				}
				if (colName.contentEquals("weight")) {
					weight = Double.parseDouble(colValue);
				}
			}
			node_list[start_node][end_node] = weight;
			node_list[end_node][start_node] = weight; // Undirected.

		}

		transport.flush();
		transport.close();

		// Print out what we've got.		
		System.out.print("\t");
		for (int i = 0; i < high_node; ++i) {
			System.out.print("\t" + (i+1) + ":");
		}
		System.out.print("\n");

		for (int i = 0; i < node_list.length; ++i ) {
			System.out.print((i + 1) + "\t|");
			for (int j = 0; j < node_list.length; ++j) {
				System.out.print("\t" + node_list[i][j]);
			}
			System.out.print("\n");
		}

		// Calculate the transitive closure.

		/*
		 * 0 = no path.
		 */
		for (int k = 0; k < high_node; ++k) {
			for (int i = 0; i < high_node; ++i) {
				for (int j = 0; j < high_node; ++j) {
					if ((node_list[i][k] != 0.0) && (node_list[k][j] != 0)) { // This is a path from [i][k] to [k][j]
						if (((node_list[i][k] + node_list[k][j]) < node_list[i][j]) || (node_list[i][j] == 0.0))
							node_list[i][j] = node_list[i][k] + node_list[k][j];
					}
				}
			}
		}

		// Print out the results of the closure.
		System.out.print("\n\t");
		for (int i = 0; i < high_node; ++i) {
			System.out.print("\t" + (i + 1) + ":");
		}
		System.out.print("\n");

		for (int i = 0; i < node_list.length; ++i ) {
			System.out.print((i+1) + "\t|");
			for (int j = 0; j < node_list.length; ++j) {
				System.out.print("\t" + node_list[i][j]);
			}
			System.out.print("\n");
		}


	}
}
